from itertools import groupby
from typing import List

import numpy as np

from mckinseysolvegame.domain.models import Species, OptimizationResult


class Solver:

    @staticmethod
    def find_sustainable_food_chain(species: List[Species]) -> OptimizationResult:
        # Check that input is consistent
        # ...

        if not species:
            return OptimizationResult(0, [])

        # Find sustainable food chain
        species = Solver.get_optimal_species_group(species)
        n = len(species)
        species.sort(key=lambda x: x.calories_provided, reverse=True)
        dp = [1] * n
        p = [-1] * n
        all_chains = [[] for _ in range(n)]  # List to store all chains for each species

        for i in range(n):
            # Iterate through previous species
            for j in range(i):
                # Check if the j-th species can be a food source for the i-th species
                if species[j].name in species[i].food_sources and species[j].calories_provided > species[i].calories_needed:
                    # Update dp and predecessor array if a longer chain is found
                    if dp[j] + 1 > dp[i]:
                        dp[i] = dp[j] + 1
                        p[i] = j
            # Store the current chain in the all_chains list
            all_chains[i] = [i] + all_chains[p[i]] if p[i] != -1 else [i]

        # Initialize sets to track unique species and unique chains
        unique_species = set()
        all_chains.sort(key=lambda x: len(x), reverse=True)
        unique_chains = []

        # Find chains containing producers and no duplicate species
        producers_indices = [i for i, species in enumerate(species) if species.calories_needed == 0]
        for chain in all_chains:
            if not any(val in unique_species for val in chain) and any(idx in chain for idx in producers_indices):
                unique_chains.append(chain)
                unique_species.update(chain)

        # Create a list of selected species based on unique chains
        selected_species = [species[i] for i in [x for c in unique_chains for x in c]]

        # Sort selected species based on calories provided
        selected_species.sort(key=lambda x: x.calories_provided, reverse=True)

        return OptimizationResult(len(selected_species), [s.name for s in selected_species])
    
    @staticmethod
    def get_optimal_species_group(species: List[Species]) -> List[Species]:
        # Find the group of producers that maximize the calories provided
        producers = [s for s in species if s.calories_needed == 0]
        grouped_producers = {key: list(group) for key, group in groupby(producers, key=lambda x: x.depth_range)}

        total_calories_per_depth_range = {}
        for depth_range, producers_group in grouped_producers.items():
            total_calories = sum(species.calories_provided for species in producers_group)
            total_calories_per_depth_range[depth_range] = total_calories

        # Find the depth range with the highest total calories
        optimal_depth_range = max(total_calories_per_depth_range, key=total_calories_per_depth_range.get)

        # Restrict the list of species to the list of species with matching depth range
        species = [s for s in species if s.depth_range == optimal_depth_range]

        return species



import json
from typing import List

from marshmallow import Schema, fields, post_load, validate

MAXIMUM_NUMBER_OF_CHARACTERS_SPECIES_NAME = 64


class Species:
    def __init__(self, name: str,
                 calories_provided: int,
                 calories_needed: int,
                 depth_range: str,
                 temperature_range: str,
                 food_sources: List[str]):
        assert len(name) <= MAXIMUM_NUMBER_OF_CHARACTERS_SPECIES_NAME, \
            f'name should be {MAXIMUM_NUMBER_OF_CHARACTERS_SPECIES_NAME} characters or less'
        self.name = name
        assert calories_provided >= 0, 'calories provided must be a positive integer'
        self.calories_provided = calories_provided
        assert calories_needed >= 0, 'calories needed must be a positive integer'
        self.calories_needed = calories_needed
        self.depth_range = depth_range
        self.temperature_range = temperature_range
        self.food_sources = food_sources

    @classmethod
    def from_json(cls, data, many=False):
        schema = SpeciesSchema(many=many)
        if isinstance(data, dict):
            return schema.load(data)
        elif isinstance(data, str):
            json_data = json.loads(data)
            return schema.load(json_data)
        else:
            raise NotImplementedError(f"Invalid input type \'{type(data)}\' during deserialization")

    def to_json(self, format_type: str = "dict"):
        schema = SpeciesSchema()
        my_json = schema.dump(self)
        if format_type == "dict":
            return my_json
        elif format_type == "str":
            return json.dumps(my_json)
        else:
            raise NotImplementedError(f"Invalid format type \'{format_type}\' during serialization")


class SpeciesSchema(Schema):
    """
        A marshmallow schema for the Species class.
    """
    name = fields.Str(required=True, validate=validate.Length(max=MAXIMUM_NUMBER_OF_CHARACTERS_SPECIES_NAME),
                      allow_none=False)
    calories_provided = fields.Int(required=True, validate=validate.Range(min=1), allow_none=False)
    calories_needed = fields.Int(validate=validate.Range(min=0), allow_none=False)
    depth_range = fields.Str(required=True, allow_none=False)
    temperature_range = fields.Str(required=True, allow_none=False)
    food_sources = fields.List(fields.Str(validate=validate.Length(max=MAXIMUM_NUMBER_OF_CHARACTERS_SPECIES_NAME)), required=True)

    @post_load
    def make_species(self, data, **kwargs) -> Species:
        return Species(**data)


class OptimizationResult:
    """
        Class representing the result of the sustainable chain of species.

        Attributes:
            number_of_species (int): The total number of species in the sustainable food chain.
            species (List[str]): The list of species names ordered by their calories provided.
    """
    def __init__(self, number_of_species: int, species: List[str]):
        self.number_of_species = number_of_species
        self.species = species

    @classmethod
    def from_json(cls, data, many=False):
        schema = OptimizationResultSchema(many=many)
        if isinstance(data, dict):
            return schema.load(data)
        elif isinstance(data, str):
            json_data = json.loads(data)
            return schema.load(json_data)
        else:
            raise NotImplementedError(f"Invalid input type \'{type(data)}\' during deserialization")

    def to_json(self, format_type: str = "dict"):
        schema = OptimizationResultSchema()
        my_json = schema.dump(self)
        if format_type == "dict":
            return my_json
        elif format_type == "str":
            return json.dumps(my_json)
        else:
            raise NotImplementedError(f"Invalid format type \'{format_type}\' during serialization")


class OptimizationResultSchema(Schema):
    """
        Marshmallow schema for deserializing and serializing OptimizationResult objects.
    """
    number_of_species = fields.Integer(required=True)
    species = fields.List(fields.String(), required=True)

    @post_load
    def make_optimization_result(self, data, **kwargs) -> OptimizationResult:
        return OptimizationResult(**data)
